use core::{
    cell::UnsafeCell,
    mem::ManuallyDrop,
    ops::{Deref, DerefMut},
};

/// A guard that runs a closure when it is dropped.
pub struct DropGuard<F: FnOnce()>(UnsafeCell<ManuallyDrop<F>>);

impl<F> DropGuard<F>
where
    F: FnOnce(),
{
    pub fn new(f: F) -> DropGuard<F> {
        Self(UnsafeCell::new(ManuallyDrop::new(f)))
    }

    pub fn guard<T>(self, t: T) -> DropGuarded<T, F> {
        DropGuarded(t, self)
    }
}

impl<F> Drop for DropGuard<F>
where
    F: FnOnce(),
{
    fn drop(&mut self) {
        // SAFETY: drop() is called exactly once.
        unsafe {
            ManuallyDrop::take(&mut *self.0.get())();
        }
    }
}

impl<F> From<F> for DropGuard<F>
where
    F: FnOnce(),
{
    fn from(f: F) -> Self {
        DropGuard::new(f)
    }
}

pub struct DropGuarded<T, F: FnOnce()>(T, DropGuard<F>);

impl<T, F> DropGuarded<T, F>
where
    F: FnOnce(),
{
    pub fn new(value: T, f: F) -> Self {
        Self(value, DropGuard::new(f))
    }

    pub fn into_inner(self) -> T {
        self.0
    }

    pub fn map<U, G>(self, f: G) -> DropGuarded<U, F>
    where
        G: FnOnce(T) -> U,
    {
        DropGuarded(f(self.0), self.1)
    }
}

impl<T, F> Deref for DropGuarded<T, F>
where
    F: FnOnce(),
{
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl<T, F> DerefMut for DropGuarded<T, F>
where
    F: FnOnce(),
{
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}
